import logging

from django.conf import settings
from django.http import HttpResponseRedirect, HttpResponse
from django.urls import path, include
from rest_framework.decorators import api_view
from rest_framework.decorators import authentication_classes
from rest_framework.decorators import permission_classes
from rest_framework.permissions import AllowAny
from cognito_code_grant.helpers import get_cookie_domain

import requests

TOKEN_TYPES = ['id_token', 'access_token', 'refresh_token']

logger = logging.getLogger(__package__)


@api_view(['GET'])
@authentication_classes([])
@permission_classes([AllowAny])
def login(request):
    auth_redirect_url: str = request.build_absolute_uri().split('?')[0]
    # since the app is running in container, no way to know its on ssl
    auth_redirect_url = auth_redirect_url.replace('http', 'https')
    auth_code: str = request.query_params.get('code', '')
    app_redirect_url: str = request.query_params.get('state', '')

    cognito_reply = requests.post(settings.AUTH_COGNITO_CODE_GRANT_URL, data={
        'grant_type': 'authorization_code',
        'code': auth_code,
        'client_id': settings.AUTH_COGNITO_CLIENT_ID,
        'redirect_uri': auth_redirect_url
    })

    try:
        cognito_reply.raise_for_status()
    except requests.exceptions.HTTPError as e:
        logger.warning("Auth request failed with status %s: %s",
                       e.response.status_code, e.response.content)
        return HttpResponse('Unauthorized', status=401)
    tokens: dict = cognito_reply.json()
    response = HttpResponseRedirect(app_redirect_url)
    cookie_domain = get_cookie_domain(request)
    for token_type in TOKEN_TYPES:
        request.session[token_type] = tokens[token_type]
        if getattr(settings, 'SHARED_TOKENS', None):
            if token_type == 'refresh_token':
                expiry_time = 60 * 60 * 30
            else:
                expiry_time = 60 * 60
            response.set_cookie(
                token_type, tokens[token_type], domain=cookie_domain, expires=expiry_time)
    return response


@api_view(['GET'])
@authentication_classes([])
@permission_classes([AllowAny])
def logout(request):
    app_redirect_url: str = request.query_params.get('state', '')
    response = HttpResponseRedirect(app_redirect_url)
    request.session.flush()
    if getattr(settings, 'SHARED_TOKENS', None):
        cookie_domain = get_cookie_domain(request)
        for token_type in TOKEN_TYPES:
            response.delete_cookie(token_type, domain=cookie_domain)
    return response


def include_auth_urls():
    return include([
        path(r'login/', login),
        path(r'logout/', logout)
    ])




from django.conf import settings
from django.contrib.auth.models import User, Group

from rest_framework import authentication
from rest_framework import exceptions
from jose import jwt

from django.core.cache import cache

import requests


def _verify_token_and_decode(token: str, jwks: dict = None):
    # provided keys over cached keys over fetching new keys
    jwks = jwks or cache.get('jwks', None) or _fetch_cognito_keys()

    unverified_header = jwt.get_unverified_header(token)
    rsa_key = alg = None
    for key in jwks["keys"]:
        if key["kid"] == unverified_header["kid"]:
            rsa_key = key
            alg = key['alg']

    if not rsa_key:
        cache.delete('jwks')
        raise ValueError('No correct keys found to decode the auth payload!')

    return jwt.decode(
        token,
        rsa_key,
        audience=settings.AUTH_COGNITO_CLIENT_ID,
        algorithms=[alg],
        options={'verify_at_hash': False}) # TODO: verify this claim properly, see https://stackoverflow.com/questions/30356460/how-do-i-validate-an-access-token-using-the-at-hash-claim-of-an-id-token


def _fetch_cognito_keys():
    jwks = requests.get(settings.AUTH_COGNITO_JWKS_URL).json()
    cache.set('jwks', jwks, timeout=None)
    return jwks


def _refresh_tokens(refresh_token: str):
    return requests.post(settings.AUTH_COGNITO_CODE_GRANT_URL, data={
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token,
            'client_id': settings.AUTH_COGNITO_CLIENT_ID
        }).json()


class CognitoAuthentication(authentication.BaseAuthentication):
    def get_token(self, request, token):
        session_token = request.session.get(token)
        shared_token = None
        if getattr(settings, 'SHARED_TOKENS', None):
            shared_token = request.COOKIES.get(token)
        return session_token or shared_token or None

    def set_token(self, request, token, value):
        request.session[token] = value

    def authenticate(self, request):
        access_token = self.get_token(request, 'access_token')
        id_token = self.get_token(request, 'id_token')
        refresh_token = self.get_token(request, 'refresh_token')

        if not access_token:
            # auth not attempted
            return None

        try:
            _verify_token_and_decode(access_token)
        except jwt.JWTError:
            new_tokens = _refresh_tokens(refresh_token)
            id_token = new_tokens.get('id_token')
            access_token = new_tokens.get('access_token')
            if not id_token or not access_token:
                raise exceptions.AuthenticationFailed('Failed to fetch new tokens - invalid refresh token')
        except ValueError:
            # couldnt find correct keys to decode the token, probably bad tokens
            raise exceptions.AuthenticationFailed('Failed to fetch new tokens - couldnt fetch keys')
        decoded_id_token = _verify_token_and_decode(id_token)
        user = self.set_user(decoded_id_token)

        self.set_token(request, 'id_token', id_token)
        self.set_token(request, 'access_token', access_token)

        return (user, None)

    def set_user(self, decoded_id_token: dict):
        email = decoded_id_token['email'].lower()
        user_id = decoded_id_token['cognito:username']
        user, _ = User.objects.get_or_create(username=user_id, email__iexact=email, defaults={'username': user_id, 'email': email})
        self.set_groups(user, decoded_id_token)

        return user

    def set_groups(self, user: User, decoded_id_token: dict):
        groups = decoded_id_token.get('cognito:groups', [])
        user_groups = user.groups.values_list('name', flat=True)
        missing_groups = set(groups) - set(user_groups)

        for missing_group in missing_groups:
            group, created = Group.objects.get_or_create(name=missing_group)
            group.user_set.add(user)



from django.conf import settings

from datetime import datetime
import time
import logging
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import padding
from botocore.signers import CloudFrontSigner
from urllib.parse import urlparse

from cognito_code_grant.helpers import get_assets_v2_domain


# Heavily inspired on a couple of existing solutions in stackoverflow and AWS documentation.
# Adapted for good measure, to serve our needs.

def rsa_signer(message):
    return private_key.sign(message, padding.PKCS1v15(), hashes.SHA1())


def generate_signed_cookies(resource=None,expire_minutes=30, assets_domain=None):
    """
    @resource   path to s3 object inside bucket(or a wildcard path,e.g. '/blah/*' or  '*')
    @expire_minutes     how many minutes before we expire these access credentials (within cookie)
    return tuple of domain used in resource URL & dict of name=>value cookies
    """
    if not resource:
        resource = '*'

    dist = SignedCookiedCloudfrontDistribution(assets_domain)
    return dist.create_signed_cookies(resource,expire_minutes=expire_minutes)


class SignedCookiedCloudfrontDistribution():

    def __init__(self,cname):
        self.domain = cname

    def get_http_resource_url(self,resource=None,secure=False):
        """
        @resource   optional path and/or filename to the resource 
                   (e.g. /mydir/somefile.txt);
                    defaults to wildcard if unset '*'
        @secure     whether to use https or http protocol for Cloudfront URL - update  
                    to match your distribution settings 
        return constructed URL
        """
        if not resource:
            resource = '*'
        protocol = "http" if not secure else "https"
        http_resource = '%s://%s/%s' % (protocol,self.domain,resource)
        return http_resource

    def create_signed_cookies(self,resource,expire_minutes=3):
        """
        generate the Cloudfront download distirbution signed cookies
        @resource   path to the file, path, or wildcard pattern to generate policy for
        @expire_minutes  number of minutes until expiration
        return      tuple with domain used within policy (so it matches 
                    cookie domain), and dict of cloudfront cookies you
                    should set in request header
        """
        http_resource = self.get_http_resource_url(resource,secure=True)    #per-file access #NOTE secure should match security settings of cloudfront distribution

        cloudfront_signer = CloudFrontSigner(settings.CLOUDFRONT_SIGNED_COOKIES_KEY_PAIR_ID, rsa_signer)
        expires = SignedCookiedCloudfrontDistribution.get_expires(expire_minutes)
        policy = cloudfront_signer.build_policy(http_resource,datetime.fromtimestamp(expires))
        encoded_policy = cloudfront_signer._url_b64encode(policy.encode('utf-8')).decode('utf-8')

        #assemble the 3 Cloudfront cookies
        signature = rsa_signer(policy.encode('utf-8'))
        encoded_signature = cloudfront_signer._url_b64encode(signature).decode('utf-8')
        cookies = {
            "CloudFront-Policy": encoded_policy,
            "CloudFront-Signature": encoded_signature,
            "CloudFront-Key-Pair-Id": settings.CLOUDFRONT_SIGNED_COOKIES_KEY_PAIR_ID,
        }
        return cookies

    @staticmethod
    def get_expires(minutes):
        unixTime = time.time() + (minutes * 60)
        expires = int(unixTime)  #if not converted to int causes Malformed Policy error and has 2 decimals in value
        return expires


if getattr(settings, 'CLOUDFRONT_SIGNED_COOKIES', False):
    #dist = SignedCookiedCloudfrontDistribution(settings.CLOUDFRONT_SIGNED_COOKIES_CNAME)
    private_key = serialization.load_pem_private_key(
        settings.CLOUDFRONT_SIGNED_COOKIES_PRIVATE_KEY.encode('utf-8'),
        password=None,
        backend=default_backend()
    )


def add_signed_cookies(response, assets_domain, domain=settings.CLOUDFRONT_SIGNED_COOKIES_DOMAIN):
    # Generate signed cookies and add them to the response.
    # The idea is to generate the cookies at each request and have them relatively short lived.
    try:
        if getattr(settings, 'CLOUDFRONT_SIGNED_COOKIES', False):
            cookies = generate_signed_cookies(settings.CLOUDFRONT_SIGNED_COOKIES_RESOURCE, assets_domain=assets_domain)
            for key, value in cookies.items():
                response.set_cookie(key, value, domain=domain)
    except Exception as ex:
        logging.error(ex)


def get_domain_from_header(host_header):
    parsed_header = urlparse(host_header).netloc or urlparse(host_header).path
    domain = '.'.join(parsed_header.split('.')[-2:])
    return '.' + domain

class SignedCookiesMiddleware(object):
    def __init__(self, get_response=None):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        if getattr(request, 'user', None) and request.user.is_authenticated:
            if 'HTTP_HOST' in request.META:
                host = request.META['HTTP_HOST']
                add_signed_cookies(response, get_assets_v2_domain(request), get_domain_from_header(host))
            else:
                add_signed_cookies(response, 'assets-v2.stitchdesignlab.com')

        return response

